import json
import logging
import os
import socket
import subprocess
import sys
from threading import Thread
from time import time, sleep
from datetime import datetime

import requests
from flask import Flask, request
from flask_script import Manager, Server

log = logging.getLogger('werkzeug')
log.setLevel(logging.ERROR)

cado_master = os.environ["CADO_MASTER"]
avx = os.environ["AVX"]
api_url = os.environ["API_URL"]
yafu_debug = False
yafu_test_seconds = 31
yafu_running = False

VERSION = "2.5"

yafu_started_at = -1

filtered_candidates = []
current_candidate = ""
current_candidate_segment = 1
current_block = 0
job = ""
stop_yafu = False
ip = ""

def refreshYafuDuration():
    global yafu_test_seconds
    global api_url

    while True:
        try:
            url = "http://" + api_url + "/requirements"
            response = requests.get(url)
            response_json = response.json()

            if response_json['yafu0Duration'] and response_json['yafu0Duration'] != yafu_test_seconds:
                print("updating yafu test duration to: ", response_json['yafu0Duration'], "sek")
                yafu_test_seconds = response_json['yafu0Duration']

        except Exception as e:
            print(e)
            print("could not load pool settings...will continue")

        sleep(60)


def custom_call():
    global VERSION
    global ip
    global cado_master
    global avx

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect(("8.8.8.8", 80))
    ip = s.getsockname()[0]
    s.close()

    t = Thread(target=refreshYafuDuration, args=())
    t.start()

    print("init yafu ... please wait ...")
    p2 = subprocess.run(["sh", "yafurun.sh", "9", avx], capture_output=True, encoding="utf-8")
    output = p2.stdout

    if "P1 = 3" not in output:
        print("could not detect yafurun.sh or it is not running correctly!!!")
        exit()

    print ("yafu init success")

    print("time:", time(), "CLIENT startup version ", VERSION, " time: ", time(), "ip: ", ip, " cado_master:", cado_master)
    pass

def checkRuntime():
    global yafu_running
    global yafu_started_at
    global candidate
    global yafu_test_seconds
    global current_block
    global candidate_level
    global candidate_block
    global filtered_candidates
    global current_candidate
    global current_candidate_segment
    global stop_yafu
    global VERSION
    global avx

    while True:
        if yafu_running == False:
            sleep(0.5)
            continue

        duration = time() - yafu_started_at

        print("yafu running since: ", duration, "  sec. ", end="\r", flush=True)

        if duration > yafu_test_seconds:
            print()
            print()
            print("adding good candidate: ", current_candidate, ", block: ", current_block, ", segment: ", current_candidate_segment)

            try:
                payload = {'height': current_block, 'cands': [current_candidate], 'host': socket.gethostname(), "duration": duration, "version": VERSION, "ip": ip, "segment": current_candidate_segment}
                url = "http://" + cado_master + "/submitfilteredchunk"
                x = requests.post(url, json=payload)
            except:
                print("failed to submit candidate to cado master. will continue ...")

            filtered_candidates.append(current_candidate)

            subprocess.run(["killall", "yafu"])
            subprocess.run(["killall", "yafu512"])
            yafu_running = False

        try:
            url = "http://" + cado_master + "/shouldcancel?job=" + job
            should_cancel = requests.get(url)

            if str(should_cancel.text) == "true":
                print()
                print()
                print("master requests to kill yafu now!")
                print()
                print()
                stop_yafu = True
                subprocess.run(["killall", "yafu"])
                subprocess.run(["killall", "yafu512"])
                yafu_running = False

        except Exception as e:
            print("failed to get job state. Error was: ", e)

        sleep(0.5)

def startYafu(candidate):
    global yafu_running
    global yafu_started_at
    global avx

    if yafu_running:
        return

    if candidate == "":
        return

    yafu_running = True
    yafu_started_at = time()

    date_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
    print("time: ", date_time, "next candidate ", candidate)
    subprocess.run(["killall", "yafu"])
    subprocess.run(["killall", "yafu512"])
    p2 = subprocess.run(["sh", "yafurun.sh", str(candidate), avx], capture_output=True, encoding="utf-8")

    if yafu_debug is True:
        print(p2.stdout)
        print(p2.stderr)

    print("yafu ended")

    yafu_running = False
    yafu_started_at = -1

app = Flask(__name__)

@app.route("/submitchunk", methods=['POST'])
def submitChunk():
    global current_candidate
    global current_candidate_segment
    global filtered_candidates
    global current_block
    global job
    global stop_yafu

    filtered_candidates = []

    data = request.get_json()

    candidates = data["chunk"]
    block = data["height"]
    segment = data["segment"]

    subprocess.run(["killall", "msieve"])
    print("received chunk with ", len(candidates), ", block: ", block, "segment: ", segment)

    current_block = block
    job = request.args.get('job')
    stop_yafu = False

    i = 1
    for candidate in candidates:

        if stop_yafu:
            break

        while yafu_running:
            sleep(0.5)

        current_candidate = candidate
        current_candidate_segment = segment
        startYafu(candidate)

        date_time = datetime.now().strftime("%m/%d/%Y, %H:%M:%S")
        print("time: ", date_time, "processed candidates: ", i, "/", len(candidates))
        i += 1

    print("finished yafu filter. received: ", len(candidates), " - remaining candidates after filter: ", len(filtered_candidates))

    #submit chunk back to cado-master
    payload = {'height': str(block), 'job': job, 'host': socket.gethostname(), "ip": ip}
    url = "http://" + cado_master + "/submitfinishedjob"

    try:
        x = requests.post(url, json=payload)
    except:
        print("failed to report finished job...will continue")

    return "ok"

class CustomServer(Server):
    def __call__(self, app, *args, **kwargs):
        custom_call()

        t = Thread(target=checkRuntime, args=())
        t.start()

        #Hint: Here you could manipulate app
        return Server.__call__(self, app, *args, **kwargs)

manager = Manager(app)
app.returned_candidates = set()

# Remeber to add the command to your Manager instance
manager.add_command('runserver', CustomServer())

if __name__ == "__main__":
    manager.run()
